Package: hypertext
Title: 'HTML' Element Construction
Version: 1.3.0
Authors@R: 
    c(
      person("Kennedy", "Mwavu", , "mwavukennedy@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0009-0006-3157-7234")),
      person("Sigflux", role = c("cph", "fnd"))
    )
Description: Provides a deterministic, framework-agnostic Domain-Specific Language for building
    'HTML' nodes and rendering them to a string.
License: MIT + file LICENSE
URL: https://sigflux.github.io/hypertext/,
        https://github.com/sigflux/hypertext
BugReports: https://github.com/sigflux/hypertext/issues
Encoding: UTF-8
Suggests: testthat (>= 3.0.0)
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
Config/Needs/website: pkgdown
NeedsCompilation: no
Packaged: 2026-06-19 18:55:55 UTC; mwavu
Author: Kennedy Mwavu [aut, cre] (ORCID:
    <https://orcid.org/0009-0006-3157-7234>),
  Sigflux [cph, fnd]
Maintainer: Kennedy Mwavu <mwavukennedy@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-21 14:40:13 UTC
Built: R 4.6.0; ; 2026-06-21 15:37:50 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,hypertext.raw)
S3method(print,hypertext.tag)
S3method(print,hypertext.tag.list)
S3method(render,default)
S3method(render,hypertext.raw)
S3method(render,hypertext.tag)
S3method(render,list)
export(doctype)
export(raw_html)
export(render)
export(tag)
export(tag_list)
export(tags)

# hypertext

<!-- badges: start -->

[![R-CMD-check](https://github.com/sigflux/hypertext/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/sigflux/hypertext/actions/workflows/R-CMD-check.yaml)
[![CRAN status](https://www.r-pkg.org/badges/version/hypertext)](https://CRAN.R-project.org/package=hypertext)
[![Lifecycle: stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)

<!-- badges: end -->

html element construction for R.

`hypertext` provides a deterministic, framework-agnostic DSL for building
html nodes and rendering them to a string.

it does **not** implement templating, dependency management, widgets
or framework integrations.

heavily inspired by [{htmltools}](https://github.com/rstudio/htmltools).

## installation

install the stable version from CRAN:

```r
install.packages("hypertext")
```

or get the development version from github:

```r
devtools::install_github("sigflux/hypertext")
```

## quick start

```r
library(hypertext)

page <- tag_list(
  doctype(),
  tags$html(
    tags$head(
      tags$title("hypertext")
    ),
    tags$body(
      tags$h1("Hello"),
      tags$p(
        class = c("lead", "mb-2"),
        "Server-side HTML."
      ),
      tags$input(
        type = "text",
        placeholder = "enter your nickname"
      ),
      tags$button(
        "Click"
      )
    )
  )
)

render(page)
```

## rendering

`render()` takes a tag tree and returns a single HTML string:

```r
x <- tags$p(
  class = "lead",
  "hello"
)

# `x` contains the tag tree
class(x)
#> [1] "hypertext.tag"

# rendering produces an HTML string:
render(x)
#> [1] "<p class=\"lead\">hello</p>"
```

you can `render()` directly to an html file by supplying
the `file` parameter:

```r
library(hypertext)

page <- tag_list(
  doctype(),
  tags$html(
    tags$head(
      tags$title("hypertext")
    ),
    tags$body(
      tags$h1("Hello"),
      tags$p(
        class = c("lead", "mb-2"),
        "Server-side HTML."
      ),
      tags$input(
        type = "text",
        placeholder = "enter your nickname"
      ),
      tags$button(
        "Click"
      )
    )
  )
)

render(x = page, file = "index.html")
```

## tag lists

`tag_list()` groups sibling nodes without wrapping them in a parent element.

```r
library(hypertext)

header <- tag_list(
  tags$h1("hello"),
  tags$p(
    class = "lead",
    "welcome aboard."
  )
)

render(header)
#> [1] "<h1>hello</h1><p class=\"lead\">welcome aboard.</p>"
```

## raw html

`raw_html()` marks a string as pre-rendered HTML so that `render()` outputs
it verbatim, without escaping. useful for injecting inline scripts, styles,
SVG markup, or any content that is already valid HTML.

```r
library(hypertext)

page <- tags$div(
  raw_html("<svg viewBox='0 0 100 100'><circle cx='50' cy='50' r='40'/></svg>")
)

render(page)
#> [1] "<div><svg viewBox='0 0 100 100'><circle cx='50' cy='50' r='40'/></svg></div>"
```

`doctype()` is a convenience wrapper around `raw_html("<!DOCTYPE html>")`.

## custom tags

`tag()` creates elements for any tag name, including web components and
custom elements not in the built-in `tags` list. it takes 3 arguments:

- `tag_name`: name of the html element.
- `...`: attributes (named) & children (unnamed).
- `tag_type`: either "normal" (default) for the standard html
  elements, or "void" for the self-closing elements.

```r
library(hypertext)

content <- tag(
  tag_name = "calcite-action-bar",
  layout = "horizontal"
)

render(content)
#> [1] "<calcite-action-bar layout=\"horizontal\"></calcite-action-bar>"
```

nest them freely with each other and with built-in tags:

```r
page <- tags$div(
  class = "app",
  tag(
    tag_name = "calcite-shell",
    tag(
      tag_name = "calcite-shell-panel",
      slot = "panel-start",
      tag(
        tag_name = "calcite-action-bar",
        tag(
          tag_name = "calcite-action",
          text = "Layers",
          icon = "layers"
        ),
        tag(
          tag_name = "calcite-action",
          text = "Basemaps",
          icon = "basemap"
        )
      )
    ),
    tags$div(id = "map")
  )
)

render(page)
```

for self-closing elements, set `tag_type = "void"`:

```r
content <- tag(
  tag_name = "my-icon",
  name = "home",
  tag_type = "void"
)

render(content)
#> [1] "<my-icon name=\"home\" />"
```

## usage in frameworks

- [ambiorix](https://ambiorix.dev/) is the perfect example of
  a web framework where you will find {hypertext} useful:

  ```r
  library(ambiorix)
  library(hypertext)

  app <- Ambiorix$new(port = 3000L)

  app$get("/", function(req, res) {
    html <- tags$h1("hello, world!") |>
      render()

    res$send(html)
  })

  app$get("/about", function(req, res) {
    html <- tag_list(
      tags$h1("about us"),
      tags$p(
        "minimal ",
        tags$strong("html construction"),
        " for R."
      )
    ) |>
      render()

    res$send(html)
  })

  app$get("/team", function(req, res) {
    teammates <- c("you", "me", "other")

    html <- tags$div(
      class = "team",
      tags$p("meet the team:"),
      tags$ul(
        lapply(teammates, tags$li)
      )
    ) |>
      render()

    res$send(html)
  })

  app$start()
  ```

- [shiny](https://shiny.posit.co/) already has {htmltools} tags
  internally, so you do not need {hypertext} in your shiny apps,
  but in case you do:

  ```r
  library(shiny)
  library(bslib)
  library(hypertext)

  # use `hypertext::tags` explicitly to avoid clashing with `shiny::tags`.
  ht <- hypertext::tags

  card <- function(title, body) {
    ht$div(
      class = "card mt-3",
      ht$div(
        class = "card-header",
        title
      ),
      ht$div(
        class = "card-body",
        ht$p(
          class = "card-text",
          body
        )
      )
    )
  }

  content <- ht$div(
    class = "container py-4",
    card("First card", "Some quick example text."),
    card("Second card", "Another body of text.")
  ) |>
    render()

  ui <- page(
    theme = bs_theme(version = 5L),
    # hypertext renders an HTML string, so wrap in shiny::HTML()
    HTML(content)
  )

  server <- function(input, output, session) {}

  shinyApp(ui, server)
  ```

# hypertext

<!-- badges: start -->

[![R-CMD-check](https://github.com/sigflux/hypertext/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/sigflux/hypertext/actions/workflows/R-CMD-check.yaml)
[![CRAN status](https://img.shields.io/cran/v/hypertext.svg)](https://CRAN.R-project.org/package=hypertext)
[![Lifecycle: stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)

<!-- badges: end -->

html element construction for R.

`hypertext` provides a deterministic, framework-agnostic DSL for building html nodes and rendering them to a string.

- dependency-free. zero package imports. lightweight is the right weight, as suggested in the [tinyverse](https://www.tinyverse.org/) philosophy.
- does **not** implement templating, dependency management, or widgets/framework integrations.
- heavily inspired by [{htmltools}](https://github.com/rstudio/htmltools).

## installation

install the stable version from CRAN:

```r
install.packages("hypertext")
```

or get the development version from github:

```r
devtools::install_github("sigflux/hypertext")
```

## quick start

```r
page <- tag_list(
  doctype(),
  tags$html(
    tags$head(
      tags$title("hypertext")
    ),
    tags$body(
      tags$h1("Hello"),
      tags$p(
        class = c("lead", "mb-2"),
        "Server-side HTML."
      ),
      tags$input(
        type = "text",
        placeholder = "enter your nickname"
      ),
      tags$button(
        "Click"
      )
    )
  )
)

page
#> <!DOCTYPE html>
#> <html>
#>   <head>
#>     <title>hypertext</title>
#>   </head>
#>   <body>
#>     <h1>Hello</h1>
#>     <p class="lead mb-2">Server-side HTML.</p>
#>     <input type="text" placeholder="enter your nickname" />
#>     <button>Click</button>
#>   </body>
#> </html>
```

## rendering

`render()` takes a tag tree and returns a single HTML string:

```r
x <- tags$p(
  class = "lead",
  "hello"
)

# `x` contains the tag tree
class(x)
#> [1] "hypertext.tag"
x
#> <p class="lead">hello</p>

# rendering produces an HTML string:
render(x)
#> [1] "<p class=\"lead\">hello</p>"
```

you can `render()` directly to an html file by supplying the `file` parameter:

```r
page <- tag_list(
  doctype(),
  tags$html(
    tags$head(
      tags$title("hypertext")
    ),
    tags$body(
      tags$h1("Hello"),
      tags$p(
        class = c("lead", "mb-2"),
        "Server-side HTML."
      ),
      tags$input(
        type = "text",
        placeholder = "enter your nickname"
      ),
      tags$button(
        "Click"
      )
    )
  )
)

render(x = page, file = "index.html")
```

## tag lists

`tag_list()` groups sibling nodes without wrapping them in a parent element.

```r
header <- tag_list(
  tags$h1("hello"),
  tags$p(
    class = "lead",
    "welcome aboard."
  )
)

# `header` is a "hypertext.tag.list" object:
class(header)
#> [1] "hypertext.tag.list" "list"

header
#> <h1>hello</h1>
#> <p class="lead">welcome aboard.</p>
```

## raw html

`raw_html()` marks a string as pre-rendered HTML so that `render()` outputs it verbatim, without escaping. useful for injecting inline scripts, styles, SVG markup, or any content that is already valid HTML.

```r
tags$div(
  raw_html("<svg viewBox='0 0 100 100'><circle cx='50' cy='50' r='40'/></svg>")
)
#> <div><svg viewBox='0 0 100 100'><circle cx='50' cy='50' r='40'/></svg></div>
```

`doctype()` is a convenience wrapper around `raw_html("<!DOCTYPE html>")`.

## custom tags

`tag()` creates elements for any tag name, including web components and custom elements not in the built-in `tags` list. it takes 3 arguments:

- `tag_name`: name of the html element.
- `...`: attributes (named) & children (unnamed).
- `tag_type`: either "normal" (default) for the standard html elements, or "void" for the self-closing elements.

```r
content <- tag(
  tag_name = "calcite-action-bar",
  layout = "horizontal"
)

content
#> <calcite-action-bar layout="horizontal"></calcite-action-bar>
```

nest them freely with each other and with built-in tags:

```r
page <- tags$div(
  class = "app",
  tag(
    tag_name = "calcite-shell",
    tag(
      tag_name = "calcite-shell-panel",
      slot = "panel-start",
      tag(
        tag_name = "calcite-action-bar",
        tag(
          tag_name = "calcite-action",
          text = "Layers",
          icon = "layers"
        ),
        tag(
          tag_name = "calcite-action",
          text = "Basemaps",
          icon = "basemap"
        )
      )
    ),
    tags$div(id = "map")
  )
)

page
#> <div class="app">
#>   <calcite-shell>
#>     <calcite-shell-panel slot="panel-start">
#>       <calcite-action-bar>
#>         <calcite-action text="Layers" icon="layers"></calcite-action>
#>         <calcite-action text="Basemaps" icon="basemap"></calcite-action>
#>       </calcite-action-bar>
#>     </calcite-shell-panel>
#>     <div id="map"></div>
#>   </calcite-shell>
#> </div>
```

for self-closing elements, set `tag_type = "void"`:

```r
content <- tag(
  tag_name = "my-icon",
  name = "home",
  tag_type = "void"
)

content
#> <my-icon name="home" />
```

## usage in frameworks

### ambiorix

[ambiorix](https://ambiorix.dev/) is the perfect example of a web framework where you will find {hypertext} useful:

```r
library(ambiorix)
library(hypertext)

app <- Ambiorix$new(port = 3000L)

app$get("/", function(req, res) {
  html <- tags$h1("hello, world!") |>
    render()

  res$send(html)
})

app$get("/about", function(req, res) {
  html <- tag_list(
    tags$h1("about us"),
    tags$p(
      "minimal ",
      tags$strong("html construction"),
      " for R."
    )
  ) |>
    render()

  res$send(html)
})

app$get("/team", function(req, res) {
  teammates <- c("you", "me", "other")

  html <- tags$div(
    class = "team",
    tags$p("meet the team:"),
    tags$ul(
      lapply(teammates, tags$li)
    )
  ) |>
    render()

  res$send(html)
})

app$start()
```

### shiny

[shiny](https://shiny.posit.co/) already has {htmltools} tags internally, so you do not need {hypertext} in your shiny apps, but in case you do:

```r
library(shiny)
library(bslib)
library(hypertext)

# use `hypertext::tags` explicitly to avoid clashing with `shiny::tags`.
ht <- hypertext::tags

card <- function(title, body) {
  ht$div(
    class = "card mt-3",
    ht$div(
      class = "card-header",
      title
    ),
    ht$div(
      class = "card-body",
      ht$p(
        class = "card-text",
        body
      )
    )
  )
}

content <- ht$div(
  class = "container py-4",
  card("First card", "Some quick example text."),
  card("Second card", "Another body of text.")
) |>
  render()

ui <- page(
  theme = bs_theme(version = 5L),
  # hypertext renders an HTML string, so wrap in shiny::HTML()
  HTML(content)
)

server <- function(input, output, session) {}

shinyApp(ui, server)
```

#' A Variation of `base::list()`
#'
#' @param ... Objects /// Optional.
#'            These are the elements to be contained by the list.
#'            Can be named, unnamed, or a mix of the two.
#'
#' @return List with `...` as elements.
#'
#' @keywords internal
#'
#' @noRd
.list2 <- function(...) {
  ellipsis_length <- ...length()
  ellipsis_names <- ...names()

  if (identical(ellipsis_length, 0L)) {
    return(list())
  }

  ellipsis_parse_tree <- substitute(
    expr = list(...)
  )

  has_trailing_comma <- identical(
    ellipsis_parse_tree[[length(ellipsis_parse_tree)]],
    quote(expr = )
  )

  if (has_trailing_comma) {
    ellipsis_length <- ellipsis_length - 1L
    ellipsis_names <- ellipsis_names[seq_len(ellipsis_length)]
  }

  if (identical(ellipsis_length, 0L)) {
    return(list())
  }

  out <- vector(mode = "list", length = ellipsis_length)

  for (idx in seq_len(ellipsis_length)) {
    value <- ...elt(idx)

    if (is.null(value)) {
      out[idx] <- list(NULL)
      next
    }

    out[[idx]] <- value
  }

  names(out) <- ellipsis_names

  out
}

# -- void (self-closing) elements ------------------------------------
.void_elements <- c(
  "area",
  "base",
  "br",
  "col",
  "embed",
  "hr",
  "img",
  "input",
  "link",
  "meta",
  "param",
  "source",
  "track",
  "wbr"
)

# -- html escaping ---------------------------------------------------

#' Escape special HTML characters
#'
#' Replaces `&`, `<`, `>`, `"`, and `'` with their HTML entity equivalents.
#'
#' @param x String /// Required.
#'
#' @return String with HTML special characters escaped.
#'
#' @keywords internal
#'
#' @noRd
.escape_html <- function(x) {
  x <- gsub("&", "&amp;", x, fixed = TRUE)
  x <- gsub("<", "&lt;", x, fixed = TRUE)
  x <- gsub(">", "&gt;", x, fixed = TRUE)
  x <- gsub("\"", "&quot;", x, fixed = TRUE)
  x <- gsub("'", "&#39;", x, fixed = TRUE)
  x
}

# -- attribute rendering ---------------------------------------------

#' Render a named list of attributes to a single HTML attribute string
#'
#' - `NA` values produce boolean / valueless attributes (e.g. `disabled`).
#' - `NULL` values are silently dropped.
#' - `TRUE` produces a boolean attribute; `FALSE` drops it.
#' - Character vectors of length > 1 are collapsed with a space
#'   (handy for `class = c("a", "b")`).
#'
#' @param attrs Named list /// Required.
#'              Attribute values.
#'
#' @return String (leading space included when non-empty).
#'
#' @keywords internal
#'
#' @noRd
.render_attrs <- function(attrs) {
  if (length(attrs) == 0L) {
    return("")
  }

  parts <- character(0L)

  for (nm in names(attrs)) {
    val <- attrs[[nm]]

    # NULL -> skip
    if (is.null(val)) {
      next
    }

    # FALSE -> skip; TRUE -> boolean attribute
    if (isFALSE(val)) {
      next
    }

    if (isTRUE(val)) {
      parts <- c(parts, nm)
      next
    }

    # NA -> boolean / valueless attribute
    if (length(val) == 1L && is.na(val)) {
      parts <- c(parts, nm)
      next
    }

    # character / numeric vectors -> collapse with space, escape
    val <- paste(as.character(val), collapse = " ")
    val <- .escape_html(val)
    parts <- c(parts, paste0(nm, "=\"", val, "\""))
  }

  if (length(parts) == 0L) {
    return("")
  }

  parts <- paste(parts, collapse = " ")

  paste0(" ", parts)
}

# -- node constructors -----------------------------------------------

#' Create an HTML element node
#'
#' Low-level constructor. Named arguments become attributes; unnamed
#' arguments become children (text or nested nodes).
#'
#' @param tag_name String /// Required.
#'                 The HTML element name.
#'
#' @param ... Attributes (named) and children (unnamed) /// Optional.
#'
#' @param tag_type String /// Optional.
#'                 Either `"normal"` (default) for a standard
#'                 element with children, or `"void"` for a self-closing element
#'                 whose children are ignored.
#'
#' @return List of class `"hypertext.tag"` with components `tag`, `attrs`,
#'        `children`, and `tag_type`.
#'
#' @examples
#' # web component
#' tag(tag_name = "calcite-action-bar", layout = "horizontal")
#'
#' # custom element with children
#' tag(
#'   tag_name = "my-card",
#'   class = "shadow",
#'   tag(tag_name = "my-card-header", "Title"),
#'   tag(tag_name = "my-card-body", "Content")
#' )
#'
#' # custom void element
#' tag(tag_name = "my-icon", name = "home", tag_type = "void")
#' @export
tag <- function(
  tag_name,
  ...,
  tag_type = c("normal", "void")
) {
  tag_type <- match.arg(arg = tag_type)
  dots <- list(...)

  nms <- names(dots)
  if (is.null(nms)) {
    nms <- rep(x = "", times = length(dots))
  }

  attrs_lgl <- nzchar(nms)
  attrs <- dots[attrs_lgl]

  children <- list()
  if (identical(tag_type, "normal")) {
    # flatten any bare lists among children so users can splice
    children <- .flatten_children(
      dots[!attrs_lgl]
    )
  }

  structure(
    list(
      tag = tag_name,
      attrs = attrs,
      children = children,
      tag_type = tag_type
    ),
    class = "hypertext.tag"
  )
}

#' Create a tag list (sibling container)
#'
#' Useful when you want to collect sibling nodes into a single
#' object without wrapping them in a parent element.
#'
#' @param ... Tags, text, or other renderable objects /// Optional.
#'
#' @return A list of class `"hypertext.tag.list"`.
#'
#' @examples
#' tl <- tag_list(tags$p("one"), tags$p("two"))
#' render(tl)
#' @export
tag_list <- function(...) {
  dots <- .list2(...)

  children <- Filter(
    f = Negate(is.null),
    x = dots
  )

  structure(
    children,
    class = c("hypertext.tag.list", "list")
  )
}

#' Mark a string as raw HTML
#'
#' Wraps one or more character strings so that [render()] outputs them
#' verbatim, **without** escaping HTML special characters.
#'
#' This is useful for injecting pre-rendered HTML, inline
#' `<script>`/`<style>` blocks, SVG markup, or any content that should
#' not be entity-escaped.
#'
#' @param ... Character strings of raw HTML /// Optional.
#'
#' @return A character vector of class `"hypertext.raw"`.
#'
#' @examples
#' raw_html("<script>alert('hi')</script>")
#' render(tags$div(raw_html("<em>already formatted</em>")))
#' @export
raw_html <- function(...) {
  structure(
    paste(..., collapse = ""),
    class = "hypertext.raw"
  )
}

#' Render the `<!DOCTYPE html>` declaration
#'
#' Convenience wrapper around [raw_html()] that returns the
#' HTML5 document-type declaration. Useful when building a full page.
#'
#' @return A `"hypertext.raw"` object containing `<!DOCTYPE html>`.
#'
#' @examples
#' page <- tag_list(
#'   doctype(),
#'   tags$html(
#'     tags$head(tags$title("Home")),
#'     tags$body(tags$h1("Welcome"))
#'   )
#' )
#' render(page)
#' @export
doctype <- function() {
  raw_html("<!DOCTYPE html>")
}

#' Flatten children
#'
#' Recursively unpack plain lists (but not `hypertext.tag` or
#' `hypertext.tag.list` nodes) so users can pass
#' `list(tags$li("a"), tags$li("b"))` as a single child argument.
#'
#' @param x List of children /// Required.
#'
#' @return A flat list of children.
#'
#' @keywords internal
#'
#' @noRd
.flatten_children <- function(x) {
  out <- list()
  for (el in x) {
    if (is.null(el)) {
      next
    }

    if (
      is.list(el) &&
        !inherits(el, "hypertext.tag") &&
        !inherits(el, "hypertext.tag.list") &&
        !inherits(el, "hypertext.raw")
    ) {
      out <- c(out, .flatten_children(el))
      next
    }

    out <- c(out, list(el))
  }

  out
}

# -- rendering -------------------------------------------------------

.maybe_write <- function(
  html,
  file = "",
  write_mode = c("overwrite", "append")
) {
  if (identical(file, "")) {
    return(html)
  }

  write_mode <- match.arg(arg = write_mode)

  cat(
    html,
    file = file,
    append = identical(write_mode, "append")
  )

  invisible(html)
}

#' Render an HTML node tree to a character string
#'
#' Converts a `hypertext.tag` object (and all its descendants) into an HTML
#' string. Text children are escaped; nested `hypertext.tag` children are
#' rendered recursively.
#'
#' When `file` is provided, the rendered HTML is written to the specified
#' file via [cat()] and the HTML string is returned
#' invisibly.
#'
#' @param x `hypertext.tag` object, a string, or a list of these /// Required.
#'
#' @param file String /// Optional.
#'            Path to file to print to.
#'
#' @param write_mode String /// Optional.
#'                   Either "overwrite" (default) which overwrites the contents
#'                   of `file`, or "append" which appends the HTML string to
#'                   `file`.
#'
#' @param ... Further arguments passed from or to other methods.
#'
#' @return A single character string of HTML.
#'
#' @examples
#' page <- tags$html(
#'   tags$head(
#'     tags$title("Home")
#'   ),
#'   tags$body(
#'     tags$h1("Welcome")
#'   )
#' )
#'
#' # return HTML as a string:
#' render(page)
#'
#' \dontrun{
#'   # write to a file:
#'   render(page, file = "index.html")
#' }
#'
#' @export
render <- function(
  x,
  ...
) {
  UseMethod("render")
}

#' @rdname render
#' @export
render.hypertext.tag <- function(
  x,
  file = "",
  write_mode = c("overwrite", "append"),
  ...
) {
  attr_str <- .render_attrs(x$attrs)
  is_void <- identical(x$tag_type, "void") || x$tag %in% .void_elements

  if (is_void) {
    html <- paste0("<", x$tag, attr_str, " />")

    return(
      .maybe_write(html = html, file = file, write_mode = write_mode)
    )
  }

  inner <- paste(
    vapply(
      X = x$children,
      FUN = render,
      FUN.VALUE = character(1L)
    ),
    collapse = ""
  )

  html <- paste0("<", x$tag, attr_str, ">", inner, "</", x$tag, ">")

  .maybe_write(html = html, file = file, write_mode = write_mode)
}

#' @rdname render
#' @export
render.default <- function(
  x,
  file = "",
  write_mode = c("overwrite", "append"),
  ...
) {
  html <- .escape_html(as.character(x))

  .maybe_write(html = html, file = file, write_mode = write_mode)
}

#' @rdname render
#' @export
render.hypertext.raw <- function(
  x,
  file = "",
  write_mode = c("overwrite", "append"),
  ...
) {
  html <- as.character(x)

  .maybe_write(html = html, file = file, write_mode = write_mode)
}

#' @rdname render
#' @export
render.list <- function(
  x,
  file = "",
  write_mode = c("overwrite", "append"),
  ...
) {
  html <- paste(
    vapply(
      X = x,
      FUN = render,
      FUN.VALUE = character(1L)
    ),
    collapse = ""
  )

  .maybe_write(html = html, file = file, write_mode = write_mode)
}

# -- print method ----------------------------------------------------

#' @export
print.hypertext.tag <- function(
  x,
  ...
) {
  cat(render(x), "\n")
  invisible(x)
}

#' @export
print.hypertext.tag.list <- print.hypertext.tag

#' @export
print.hypertext.raw <- print.hypertext.tag

# -- tag factory helpers ---------------------------------------------

#' Create a tag function
#'
#' @param tag_name String /// Required.
#'                 Name of the tag.
#'
#' @param tag_type String /// Optional.
#'                 Either "normal" (default) for a normal HTML tag
#'                 or "void" for a self-closing HTML tag.
#'
#' @return Function.
#'
#' @keywords internal
#'
#' @noRd
.make_tag <- function(
  tag_name,
  tag_type = c("normal", "void")
) {
  force(tag_name)
  tag_type <- match.arg(arg = tag_type)

  function(...) {
    dots <- .list2(...)

    dots$tag_name <- tag_name
    dots$tag_type <- tag_type

    do.call(what = tag, args = dots)
  }
}

# -- the `tags` list -------------------------------------------------

#' HTML tag functions
#'
#' A named list of functions, one per HTML5 element. Access individual
#' tags with `tags$div(...)`, `tags$p(...)`, etc.
#'
#' Named arguments become HTML attributes; unnamed arguments become
#' child nodes or text content.
#'
#' @section Boolean / valueless attributes:
#' Pass `NA` or `TRUE` as the attribute value to produce a valueless
#' attribute (e.g. `disabled`). `FALSE` or `NULL` suppresses the
#' attribute entirely.
#'
#' @section Multi-value attributes:
#' Character vectors of length > 1 are collapsed with a space, so
#' `class = c("a", "b")` renders as `class="a b"`.
#'
#' @examples
#' tags$p(class = "lead", "Hello, world!")
#' render(tags$div(id = "app", tags$h1("Title")))
#' @export
tags <- list(
  # -- document metadata & root ------------------------------------
  html = .make_tag("html"),
  head = .make_tag("head"),
  body = .make_tag("body"),
  title = .make_tag("title"),
  base = .make_tag("base", tag_type = "void"),
  link = .make_tag("link", tag_type = "void"),
  meta = .make_tag("meta", tag_type = "void"),
  style = .make_tag("style"),
  script = .make_tag("script"),
  noscript = .make_tag("noscript"),

  # -- sectioning & landmarks --------------------------------------
  header = .make_tag("header"),
  footer = .make_tag("footer"),
  nav = .make_tag("nav"),
  main = .make_tag("main"),
  section = .make_tag("section"),
  article = .make_tag("article"),
  aside = .make_tag("aside"),
  address = .make_tag("address"),
  hgroup = .make_tag("hgroup"),
  search = .make_tag("search"),

  # -- headings ----------------------------------------------------
  h1 = .make_tag("h1"),
  h2 = .make_tag("h2"),
  h3 = .make_tag("h3"),
  h4 = .make_tag("h4"),
  h5 = .make_tag("h5"),
  h6 = .make_tag("h6"),

  # -- block text content ------------------------------------------
  div = .make_tag("div"),
  p = .make_tag("p"),
  blockquote = .make_tag("blockquote"),
  pre = .make_tag("pre"),
  figure = .make_tag("figure"),
  figcaption = .make_tag("figcaption"),
  hr = .make_tag("hr", tag_type = "void"),

  # -- inline text semantics ---------------------------------------
  a = .make_tag("a"),
  abbr = .make_tag("abbr"),
  b = .make_tag("b"),
  bdi = .make_tag("bdi"),
  bdo = .make_tag("bdo"),
  br = .make_tag("br", tag_type = "void"),
  cite = .make_tag("cite"),
  code = .make_tag("code"),
  data = .make_tag("data"),
  dfn = .make_tag("dfn"),
  em = .make_tag("em"),
  i = .make_tag("i"),
  kbd = .make_tag("kbd"),
  mark = .make_tag("mark"),
  q = .make_tag("q"),
  rp = .make_tag("rp"),
  rt = .make_tag("rt"),
  ruby = .make_tag("ruby"),
  s = .make_tag("s"),
  samp = .make_tag("samp"),
  small = .make_tag("small"),
  span = .make_tag("span"),
  strong = .make_tag("strong"),
  sub = .make_tag("sub"),
  sup = .make_tag("sup"),
  time = .make_tag("time"),
  u = .make_tag("u"),
  var = .make_tag("var"),
  wbr = .make_tag("wbr", tag_type = "void"),
  del = .make_tag("del"),
  ins = .make_tag("ins"),

  # -- lists -------------------------------------------------------
  ul = .make_tag("ul"),
  ol = .make_tag("ol"),
  li = .make_tag("li"),
  dl = .make_tag("dl"),
  dt = .make_tag("dt"),
  dd = .make_tag("dd"),
  menu = .make_tag("menu"),

  # -- tables ------------------------------------------------------
  table = .make_tag("table"),
  caption = .make_tag("caption"),
  colgroup = .make_tag("colgroup"),
  col = .make_tag("col", tag_type = "void"),
  thead = .make_tag("thead"),
  tbody = .make_tag("tbody"),
  tfoot = .make_tag("tfoot"),
  tr = .make_tag("tr"),
  th = .make_tag("th"),
  td = .make_tag("td"),

  # -- forms -------------------------------------------------------
  form = .make_tag("form"),
  fieldset = .make_tag("fieldset"),
  legend = .make_tag("legend"),
  label = .make_tag("label"),
  input = .make_tag("input", tag_type = "void"),
  button = .make_tag("button"),
  select = .make_tag("select"),
  option = .make_tag("option"),
  optgroup = .make_tag("optgroup"),
  datalist = .make_tag("datalist"),
  textarea = .make_tag("textarea"),
  output = .make_tag("output"),
  progress = .make_tag("progress"),
  meter = .make_tag("meter"),

  # -- media & embedded content ------------------------------------
  img = .make_tag("img", tag_type = "void"),
  picture = .make_tag("picture"),
  source = .make_tag("source", tag_type = "void"),
  audio = .make_tag("audio"),
  video = .make_tag("video"),
  track = .make_tag("track", tag_type = "void"),
  map = .make_tag("map"),
  area = .make_tag("area", tag_type = "void"),
  iframe = .make_tag("iframe"),
  embed = .make_tag("embed", tag_type = "void"),
  object = .make_tag("object"),
  param = .make_tag("param", tag_type = "void"),
  canvas = .make_tag("canvas"),
  svg = .make_tag("svg"),
  math = .make_tag("math"),

  # -- interactive / scripting -------------------------------------
  details = .make_tag("details"),
  summary = .make_tag("summary"),
  dialog = .make_tag("dialog"),
  template = .make_tag("template"),
  slot = .make_tag("slot")
)

# -- no arguments ----------------------------------------------------------

test_that(".list2() with no arguments returns an empty list", {
  expect_identical(
    .list2(),
    list()
  )
})

# -- trailing commas -------------------------------------------------------

test_that(".list2() throws when an argument is missing", {
  expect_error(
    .list2(,),
    regexp = "argument is missing, with no default"
  )
  expect_error(
    .list2("a", , ),
    regexp = "argument is missing, with no default"
  )
})

test_that(".list2() with a trailing comma after one unnamed arg", {
  expect_identical(
    .list2(1, ),
    list(1)
  )
})

test_that(".list2() with a trailing comma after one named arg", {
  expect_identical(
    .list2(a = 1, ),
    list(a = 1)
  )
})

test_that(".list2() with a trailing comma after multiple args", {
  expect_identical(
    .list2(1, 2, 3, ),
    list(1, 2, 3)
  )
  expect_identical(
    .list2(a = 1, b = 2, ),
    list(a = 1, b = 2)
  )
})

test_that(".list2() trailing comma with mixed named/unnamed args", {
  expect_identical(
    .list2(a = 1, "hello", b = 3, ),
    list(a = 1, "hello", b = 3)
  )
})

# -- NULL handling ---------------------------------------------------------

test_that(".list2(NULL) returns list(NULL)", {
  expect_identical(
    .list2(NULL),
    list(NULL)
  )
})

test_that(".list2() preserves NULL among other elements", {
  expect_identical(
    .list2(1, NULL, 3),
    list(1, NULL, 3)
  )
})

test_that(".list2() preserves named NULL", {
  expect_identical(
    .list2(a = NULL, ),
    list(a = NULL)
  )
  expect_identical(
    .list2(a = 1, b = NULL, c = 3, ),
    list(a = 1, b = NULL, c = 3)
  )
})

# -- unnamed arguments -----------------------------------------------------

test_that(".list2() with unnamed arguments", {
  expect_identical(
    .list2(1, 2, 3),
    list(1, 2, 3)
  )
  expect_identical(
    .list2("a", "b", "c", ),
    list("a", "b", "c")
  )
})

# -- named arguments -------------------------------------------------------

test_that(".list2() with a single named argument", {
  expect_identical(
    .list2(x = 1, ),
    list(x = 1)
  )
})

test_that(".list2() with multiple named arguments", {
  expect_identical(
    .list2(a = 1, b = "two", c = TRUE, ),
    list(a = 1, b = "two", c = TRUE)
  )
})

test_that(".list2() preserves names correctly", {
  result <- .list2(foo = 10, bar = 20)
  expect_identical(
    names(result),
    c("foo", "bar")
  )
  expect_identical(result$foo, 10)
  expect_identical(result$bar, 20)
})

# -- mixed named and unnamed arguments ------------------------------------

test_that(".list2() with mixed named and unnamed arguments", {
  expect_identical(
    .list2(a = 1, 2, b = 3, ),
    list(a = 1, 2, b = 3)
  )
})

test_that(".list2() mixed args have partially empty names", {
  result <- .list2(a = 1, 2, b = 3)
  expect_equal(
    names(result),
    c("a", "", "b")
  )
})

test_that(".list2() mixed args: unnamed first, named later", {
  expect_identical(
    .list2(1, 2, z = 3),
    list(1, 2, z = 3)
  )
})

test_that(".list2() mixed args: named first, unnamed later", {
  expect_identical(
    .list2(z = 3, 1, 2),
    list(z = 3, 1, 2)
  )
})

# -- various value types ---------------------------------------------------

test_that(".list2() with numeric values", {
  expect_identical(
    .list2(1L, 2.5, 3 + 1i),
    list(1L, 2.5, 3 + 1i)
  )
})

test_that(".list2() with character values", {
  expect_identical(
    .list2("a", "b"),
    list("a", "b")
  )
})

test_that(".list2() with logical values", {
  expect_identical(
    .list2(TRUE, FALSE, NA),
    list(TRUE, FALSE, NA)
  )
})

test_that(".list2() with vectors as elements", {
  expect_identical(
    .list2(1:3, c("a", "b")),
    list(1:3, c("a", "b"))
  )
})

test_that(".list2() with nested lists", {
  inner <- list(a = 1, b = 2)
  expect_identical(
    .list2(inner, 3),
    list(inner, 3)
  )
  expect_identical(
    .list2(x = inner),
    list(x = inner)
  )
})

test_that(".list2() with a function as an element", {
  fn <- function(x) x + 1
  expect_identical(
    .list2(fn),
    list(fn)
  )
})

test_that(".list2() with a formula as an element", {
  f <- y ~ x
  result <- .list2(f)
  expect_identical(
    .list2(f),
    list(f)
  )
})

test_that(".list2() with an environment as an element", {
  e <- new.env(parent = emptyenv())
  e$x <- 42
  expect_identical(
    .list2(e),
    list(e)
  )
})

test_that(".list2() with data.frame as an element", {
  df <- data.frame(x = 1:3, y = letters[1:3])
  expect_identical(
    .list2(df),
    list(df)
  )
})

test_that(".list2() with NA values", {
  expect_identical(
    .list2(NA),
    list(NA)
  )
  expect_identical(
    .list2(NA_real_),
    list(NA_real_)
  )
  expect_identical(
    .list2(NA_character_),
    list(NA_character_)
  )
})

# -- expressions and computed values ---------------------------------------

test_that(".list2() evaluates expressions in arguments", {
  expect_identical(
    .list2(1 + 1),
    list(1 + 1)
  )
  expect_identical(
    .list2(x = paste0("a", "b")),
    list(x = paste0("a", "b"))
  )
})

test_that(".list2() with variables as arguments", {
  a <- 10
  b <- "hello"
  expect_identical(
    .list2(a, b),
    list(a, b)
  )
  expect_identical(
    .list2(x = a, y = b),
    list(x = a, y = b)
  )
})

# -- edge cases ------------------------------------------------------------

test_that(".list2() with duplicate names", {
  result <- .list2(a = 1, a = 2)
  expected <- list(a = 1, a = 2)
  expect_identical(result, expected)
})

test_that(".list2() matches base::list() with do.call", {
  args <- list(a = 1, b = "two", c = TRUE)
  expect_identical(
    do.call(.list2, args),
    do.call(base::list, args)
  )
})

# -- deep forwarding -------------------------------------------------------

test_that(".list2() matches list() through one level of forwarding", {
  f <- function(...) {
    .list2(...)
  }
  g <- function(...) {
    list(...)
  }

  x <- 1
  y <- "a"

  expect_identical(
    f(x, y, z = x + 1),
    g(x, y, z = x + 1)
  )
})

test_that(".list2() matches list() through multiple levels of forwarding", {
  f3 <- function(...) {
    .list2(...)
  }
  f2 <- function(...) {
    f3(...)
  }
  f1 <- function(...) {
    f2(...)
  }

  g3 <- function(...) {
    list(...)
  }
  g2 <- function(...) {
    g3(...)
  }
  g1 <- function(...) {
    g2(...)
  }

  x <- 1
  y <- "a"

  expect_identical(
    f1(x, y, z = x + 1),
    g1(x, y, z = x + 1)
  )
})

test_that(".list2() deep forwarding uses the original caller environment", {
  f3 <- function(...) {
    .list2(...)
  }
  f2 <- function(...) {
    x <- "wrong"
    y <- "wrong"
    f3(...)
  }
  f1 <- function(...) {
    x <- "also wrong"
    f2(...)
  }

  g3 <- function(...) {
    list(...)
  }
  g2 <- function(...) {
    x <- "wrong"
    y <- "wrong"
    g3(...)
  }
  g1 <- function(...) {
    x <- "also wrong"
    g2(...)
  }

  local({
    x <- 10
    y <- 20

    expect_identical(
      f1(x, y, z = x + y),
      g1(x, y, z = x + y)
    )
  })
})

test_that(".list2() deep forwarding preserves names", {
  f2 <- function(...) {
    .list2(...)
  }
  f1 <- function(...) {
    f2(...)
  }

  g2 <- function(...) {
    list(...)
  }
  g1 <- function(...) {
    g2(...)
  }

  x <- 1
  y <- 2

  expect_identical(
    f1(a = x, y, c = y + 1),
    g1(a = x, y, c = y + 1)
  )
})

test_that(".list2() deep forwarding works with trailing commas", {
  f2 <- function(...) {
    .list2(...)
  }
  f1 <- function(...) {
    f2(...)
  }

  g2 <- function(...) {
    list(...)
  }
  g1 <- function(...) {
    g2(...)
  }

  x <- 1
  y <- 2

  expect_identical(
    f1(x, y, ),
    g1(x, y)
  )
  expect_identical(
    f1(a = x, b = y, ),
    g1(a = x, b = y)
  )
})

test_that(".list2() deep forwarding preserves NULL", {
  f2 <- function(...) {
    .list2(...)
  }
  f1 <- function(...) {
    f2(...)
  }

  g2 <- function(...) {
    list(...)
  }
  g1 <- function(...) {
    g2(...)
  }

  x <- NULL
  y <- 2

  expect_identical(
    f1(x, y, z = NULL),
    g1(x, y, z = NULL)
  )
})

test_that(".list2() deep forwarding evaluates expressions like list()", {
  f2 <- function(...) {
    .list2(...)
  }
  f1 <- function(...) {
    f2(...)
  }

  g2 <- function(...) {
    list(...)
  }
  g1 <- function(...) {
    g2(...)
  }

  x <- 2
  y <- 3

  expect_identical(
    f1(x + y, a = x * y, paste0("v", x)),
    g1(x + y, a = x * y, paste0("v", x))
  )
})

# -- raw_html construction ----------------------------------------------------

test_that("raw_html returns object with correct class", {
  r <- raw_html("<b>bold</b>")
  expect_s3_class(r, "hypertext.raw")
})

test_that("raw_html preserves the input string", {
  r <- raw_html("<em>hi</em>")
  expect_equal(as.character(r), "<em>hi</em>")
})

test_that("raw_html concatenates multiple arguments", {
  r <- raw_html("<b>one</b>", "<i>two</i>")
  expect_equal(as.character(r), "<b>one</b> <i>two</i>")
})

test_that("raw_html with no arguments returns empty string", {
  r <- raw_html()
  expect_s3_class(r, "hypertext.raw")
  expect_equal(as.character(r), "")
})

test_that("raw_html coerces non-character input via paste", {
  expect_equal(as.character(raw_html(42)), "42")
  expect_equal(as.character(raw_html(TRUE)), "TRUE")
})

# -- render.hypertext.raw ----------------------------------------------------

test_that("render does not escape raw_html content", {
  r <- raw_html("<script>alert('hi')</script>")
  expect_equal(render(r), "<script>alert('hi')</script>")
})

test_that("render preserves special characters in raw_html", {
  r <- raw_html('a < b & c > d "e" \'f\'')
  expect_equal(render(r), 'a < b & c > d "e" \'f\'')
})

test_that("raw_html works as a child of a tag", {
  node <- tags$div(raw_html("<em>raw</em>"))
  expect_equal(render(node), "<div><em>raw</em></div>")
})

test_that("raw_html mixes with text and tag children", {
  node <- tags$div(
    "before ",
    raw_html("<em>raw</em>"),
    " ",
    tags$strong("tag"),
    " after"
  )
  expect_equal(
    render(node),
    "<div>before <em>raw</em> <strong>tag</strong> after</div>"
  )
})

test_that("raw_html works inside tag_list", {
  tl <- tag_list(raw_html("<hr />"), tags$p("text"))
  expect_equal(render(tl), "<hr /><p>text</p>")
})

test_that("raw_html renders inside a list", {
  nodes <- list(raw_html("<!DOCTYPE html>"), tags$html())
  expect_equal(render(nodes), "<!DOCTYPE html><html></html>")
})

# -- render with file output -------------------------------------------------

test_that("render writes raw_html to a file and returns invisibly", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  r <- raw_html("<div>raw</div>")
  result <- withVisible(render(r, file = tmp))

  expect_false(result$visible)
  expect_equal(result$value, "<div>raw</div>")
  expect_equal(readLines(tmp, warn = FALSE), "<div>raw</div>")
})

test_that("render appends raw_html to a file", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  render(raw_html("<p>first</p>"), file = tmp)
  render(raw_html("<p>second</p>"), file = tmp, write_mode = "append")

  expect_equal(
    readLines(tmp, warn = FALSE),
    "<p>first</p><p>second</p>"
  )
})

# -- print method ------------------------------------------------------------

test_that("print.hypertext.raw outputs rendered content", {
  r <- raw_html("<b>bold</b>")
  expect_output(print(r), "<b>bold</b>")
})

test_that(".render_attrs returns empty string for empty list", {
  expect_equal(hypertext:::.render_attrs(list()), "")
})

test_that(".render_attrs renders a simple character attribute", {
  expect_equal(
    hypertext:::.render_attrs(list(class = "foo")),
    ' class="foo"'
  )
})

test_that(".render_attrs renders multiple attributes", {
  result <- hypertext:::.render_attrs(list(id = "main", class = "container"))
  expect_equal(result, ' id="main" class="container"')
})

test_that(".render_attrs drops NULL values silently", {
  expect_equal(
    hypertext:::.render_attrs(list(id = "x", class = NULL)),
    ' id="x"'
  )
})

test_that(".render_attrs drops FALSE values silently", {
  expect_equal(
    hypertext:::.render_attrs(list(id = "x", disabled = FALSE)),
    ' id="x"'
  )
})

test_that(".render_attrs renders TRUE as boolean attribute", {
  expect_equal(
    hypertext:::.render_attrs(list(disabled = TRUE)),
    " disabled"
  )
})

test_that(".render_attrs renders NA as boolean attribute", {
  expect_equal(
    hypertext:::.render_attrs(list(disabled = NA)),
    " disabled"
  )
})

test_that(".render_attrs collapses character vectors with space", {
  expect_equal(
    hypertext:::.render_attrs(list(class = c("a", "b", "c"))),
    ' class="a b c"'
  )
})

test_that(".render_attrs escapes special characters in values", {
  expect_equal(
    hypertext:::.render_attrs(list(title = 'say "hello"')),
    ' title="say &quot;hello&quot;"'
  )
})

test_that(".render_attrs renders numeric values", {
  expect_equal(
    hypertext:::.render_attrs(list(tabindex = 1)),
    ' tabindex="1"'
  )
})

test_that(".render_attrs returns empty string when all values are NULL", {
  expect_equal(
    hypertext:::.render_attrs(list(a = NULL, b = NULL)),
    ""
  )
})

test_that(".render_attrs returns empty string when all values are FALSE", {
  expect_equal(
    hypertext:::.render_attrs(list(a = FALSE, b = FALSE)),
    ""
  )
})

test_that(".render_attrs mixes boolean and valued attributes", {
  result <- hypertext:::.render_attrs(list(
    class = "btn",
    disabled = TRUE,
    hidden = NA,
    id = NULL
  ))
  expect_equal(result, ' class="btn" disabled hidden')
})

test_that(".render_attrs includes leading space when non-empty", {
  result <- hypertext:::.render_attrs(list(id = "x"))
  expect_true(startsWith(result, " "))
})

# -- integration tests: real-world patterns --------------------------------

test_that("full HTML document structure renders correctly", {
  doc <- tags$html(
    tags$head(
      tags$title("My Page"),
      tags$meta(charset = "utf-8"),
      tags$link(rel = "stylesheet", href = "style.css")
    ),
    tags$body(
      tags$h1("Hello"),
      tags$p("Welcome")
    )
  )
  html <- render(doc)
  expect_match(html, "^<html>")
  expect_match(html, "</html>$")
  expect_match(html, "<title>My Page</title>")
  expect_match(html, '<meta charset="utf-8" />')
  expect_match(html, '<link rel="stylesheet" href="style.css" />')
  expect_match(html, "<h1>Hello</h1>")
})

test_that("form with mixed void and non-void elements", {
  form <- tags$form(
    action = "/submit",
    method = "post",
    tags$label(`for` = "name", "Name:"),
    tags$input(type = "text", id = "name", name = "name"),
    tags$button(type = "submit", "Submit")
  )
  html <- render(form)
  expect_match(html, '<form action="/submit" method="post">')
  expect_match(html, '<label for="name">Name:</label>')
  expect_match(html, '<input type="text" id="name" name="name" />')
  expect_match(html, "<button type=\"submit\">Submit</button>")
})

test_that("list generated with lapply is properly flattened", {
  items <- c("Alpha", "Beta", "Gamma")
  ul <- tags$ul(lapply(items, function(x) tags$li(x)))
  html <- render(ul)
  expect_equal(
    html,
    "<ul><li>Alpha</li><li>Beta</li><li>Gamma</li></ul>"
  )
})

test_that("table with multiple rows and cells", {
  tbl <- tags$table(
    class = "data-table",
    tags$thead(
      tags$tr(tags$th("Name"), tags$th("Age"))
    ),
    tags$tbody(
      tags$tr(tags$td("Alice"), tags$td("30")),
      tags$tr(tags$td("Bob"), tags$td("25"))
    )
  )
  html <- render(tbl)
  expect_match(html, '<table class="data-table">')
  expect_match(html, "<th>Name</th><th>Age</th>")
  expect_match(html, "<td>Alice</td><td>30</td>")
})

test_that("XSS content is properly escaped in text and attributes", {
  node <- tags$div(
    title = '<script>alert("xss")</script>',
    "<img onerror='alert(1)' src=x>"
  )
  html <- render(node)
  # Attribute value should be escaped
  expect_false(grepl("<script>", html, fixed = TRUE))
  # Child text should be escaped
  expect_false(grepl("<img", html, fixed = TRUE))
  expect_match(html, "&lt;img")
})

test_that("deeply nested structure renders correctly", {
  node <- tags$div(
    tags$div(
      tags$div(
        tags$div(
          tags$span("deep")
        )
      )
    )
  )
  expect_equal(
    render(node),
    "<div><div><div><div><span>deep</span></div></div></div></div>"
  )
})

test_that("siblings rendered via render.list", {
  siblings <- list(
    tags$p("first"),
    tags$hr(),
    tags$p("second")
  )
  html <- render(siblings)
  expect_equal(html, "<p>first</p><hr /><p>second</p>")
})

test_that("multiple class values are collapsed", {
  node <- tags$div(class = c("container", "mx-auto", "p-4"), "content")
  expect_equal(
    render(node),
    '<div class="container mx-auto p-4">content</div>'
  )
})

test_that("boolean attributes in real form elements", {
  node <- tags$input(
    type = "checkbox",
    checked = TRUE,
    disabled = FALSE,
    required = NA
  )
  html <- render(node)
  expect_match(html, "checked")
  expect_false(grepl("disabled", html))
  expect_match(html, "required")
})

test_that("mixed numeric and string children", {
  node <- tags$p("Score: ", 42)
  html <- render(node)
  expect_equal(html, "<p>Score: 42</p>")
})

# -- doctype construction -----------------------------------------------------

test_that("doctype returns a hypertext.raw object", {
  d <- doctype()
  expect_s3_class(d, "hypertext.raw")
})

test_that("doctype renders to <!DOCTYPE html>", {
  expect_equal(render(doctype()), "<!DOCTYPE html>")
})

# -- full page usage ---------------------------------------------------------

test_that("doctype works in a full page with tag_list", {
  page <- tag_list(
    doctype(),
    tags$html(
      tags$head(tags$title("Test")),
      tags$body(tags$h1("Hello"))
    )
  )
  html <- render(page)
  expect_equal(
    html,
    "<!DOCTYPE html><html><head><title>Test</title></head><body><h1>Hello</h1></body></html>"
  )
})

test_that("doctype works in a plain list", {
  page <- list(
    doctype(),
    tags$html(tags$body("hi"))
  )
  html <- render(page)
  expect_equal(html, "<!DOCTYPE html><html><body>hi</body></html>")
})

# -- file output -------------------------------------------------------------

test_that("full page with doctype writes correctly to file", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  page <- tag_list(
    doctype(),
    tags$html(
      tags$head(tags$title("Test")),
      tags$body(tags$p("content"))
    )
  )
  render(page, file = tmp)

  expect_equal(
    readLines(tmp, warn = FALSE),
    "<!DOCTYPE html><html><head><title>Test</title></head><body><p>content</p></body></html>"
  )
})

# -- print method ------------------------------------------------------------

test_that("print.hypertext.raw works for doctype", {
  expect_output(print(doctype()), "<!DOCTYPE html>")
})

test_that("print.hypertext.tag outputs leaf HTML to console", {
  node <- tags$p("hello")
  expect_output(print(node), "<p>hello</p>")
})

test_that("print.hypertext.tag pretty prints nested HTML", {
  node <- tags$html(
    tags$head(
      tags$title("hypertext")
    ),
    tags$body(
      tags$h1("Hello"),
      tags$p(class = c("lead", "mb-2"), "Server-side HTML."),
      tags$input(type = "text", placeholder = "enter your nickname"),
      tags$button("Click")
    )
  )

  expect_equal(
    capture.output(print(node)),
    c(
      "<html>",
      "  <head>",
      "    <title>hypertext</title>",
      "  </head>",
      "  <body>",
      "    <h1>Hello</h1>",
      "    <p class=\"lead mb-2\">Server-side HTML.</p>",
      "    <input type=\"text\" placeholder=\"enter your nickname\" />",
      "    <button>Click</button>",
      "  </body>",
      "</html>"
    )
  )
})

test_that("print.hypertext.tag pretty prints mixed text and nested children", {
  node <- tags$p("Hello ", tags$strong("world"), "!")

  expect_equal(
    capture.output(print(node)),
    c(
      "<p>",
      "  Hello ",
      "  <strong>world</strong>",
      "  !",
      "</p>"
    )
  )
})

test_that("print.hypertext.tag pretty prints custom tag trees", {
  page <- tags$div(
    class = "app",
    tag(
      tag_name = "calcite-shell",
      tag(
        tag_name = "calcite-shell-panel",
        slot = "panel-start",
        tag(
          tag_name = "calcite-action-bar",
          tag(
            tag_name = "calcite-action",
            text = "Layers",
            icon = "layers"
          ),
          tag(
            tag_name = "calcite-action",
            text = "Basemaps",
            icon = "basemap"
          )
        )
      ),
      tags$div(id = "map")
    )
  )

  expect_equal(
    capture.output(print(page)),
    c(
      "<div class=\"app\">",
      "  <calcite-shell>",
      "    <calcite-shell-panel slot=\"panel-start\">",
      "      <calcite-action-bar>",
      "        <calcite-action text=\"Layers\" icon=\"layers\"></calcite-action>",
      "        <calcite-action text=\"Basemaps\" icon=\"basemap\"></calcite-action>",
      "      </calcite-action-bar>",
      "    </calcite-shell-panel>",
      "    <div id=\"map\"></div>",
      "  </calcite-shell>",
      "</div>"
    )
  )
})

test_that("print.hypertext.tag returns the node invisibly", {
  node <- tags$div("test")
  result <- withVisible(print(node))
  expect_false(result$visible)
  expect_identical(result$value, node)
})

test_that("print.hypertext.tag appends a newline without extra spaces", {
  node <- tags$span("x")
  output <- capture.output(print(node))
  expect_equal(output, "<span>x</span>")
})

test_that(".escape_html escapes ampersand", {
  expect_equal(hypertext:::.escape_html("a&b"), "a&amp;b")
})

test_that(".escape_html escapes less-than", {
  expect_equal(hypertext:::.escape_html("a<b"), "a&lt;b")
})

test_that(".escape_html escapes greater-than", {
  expect_equal(hypertext:::.escape_html("a>b"), "a&gt;b")
})

test_that(".escape_html escapes double quotes", {
  expect_equal(hypertext:::.escape_html('a"b'), "a&quot;b")
})

test_that(".escape_html escapes single quotes", {
  expect_equal(hypertext:::.escape_html("a'b"), "a&#39;b")
})

test_that(".escape_html escapes all special characters together", {
  expect_equal(
    hypertext:::.escape_html("&<>\"'"),
    "&amp;&lt;&gt;&quot;&#39;"
  )
})

test_that(".escape_html does not double-escape ampersands", {
  expect_equal(
    hypertext:::.escape_html("&amp;"),
    "&amp;amp;"
  )
})

test_that(".escape_html returns plain strings unchanged", {
  expect_equal(hypertext:::.escape_html("hello world"), "hello world")
})

test_that(".escape_html handles empty string", {
  expect_equal(hypertext:::.escape_html(""), "")
})

test_that(".escape_html handles string with only special characters", {
  expect_equal(hypertext:::.escape_html("<>"), "&lt;&gt;")
})

# -- render.hypertext.tag ---------------------------------------------------

test_that("render produces correct HTML for simple element", {
  node <- tags$div("hello")
  expect_equal(render(node), "<div>hello</div>")
})

test_that("render produces correct HTML for element with attributes", {
  node <- tags$p(class = "lead", "text")
  expect_equal(render(node), '<p class="lead">text</p>')
})

test_that("render handles empty element", {
  node <- tags$div()
  expect_equal(render(node), "<div></div>")
})

test_that("render handles void elements as self-closing", {
  node <- tags$br()
  expect_equal(render(node), "<br />")
})

test_that("render handles void element with attributes", {
  node <- tags$input(type = "text", name = "user")
  expect_equal(render(node), '<input type="text" name="user" />')
})

test_that("render handles nested elements", {
  node <- tags$div(tags$p("inner"))
  expect_equal(render(node), "<div><p>inner</p></div>")
})

test_that("render handles deeply nested elements", {
  node <- tags$div(tags$ul(tags$li("item")))
  expect_equal(render(node), "<div><ul><li>item</li></ul></div>")
})

test_that("render handles multiple children", {
  node <- tags$div(tags$p("first"), tags$p("second"))
  expect_equal(render(node), "<div><p>first</p><p>second</p></div>")
})

test_that("render handles mixed text and tag children", {
  node <- tags$p("Hello ", tags$strong("world"), "!")
  expect_equal(render(node), "<p>Hello <strong>world</strong>!</p>")
})

test_that("render escapes text children", {
  node <- tags$p("<script>alert('xss')</script>")
  expect_equal(
    render(node),
    "<p>&lt;script&gt;alert(&#39;xss&#39;)&lt;/script&gt;</p>"
  )
})

test_that("render escapes attribute values", {
  node <- tags$div(title = 'say "hi"')
  expect_equal(render(node), '<div title="say &quot;hi&quot;"></div>')
})

test_that("render handles boolean attributes", {
  node <- tags$input(type = "checkbox", checked = TRUE)
  expect_equal(render(node), '<input type="checkbox" checked />')
})

test_that("render handles NA as boolean attribute", {
  node <- tags$input(disabled = NA)
  expect_equal(render(node), "<input disabled />")
})

test_that("render drops FALSE attributes", {
  node <- tags$input(type = "text", disabled = FALSE)
  expect_equal(render(node), '<input type="text" />')
})

test_that("render drops NULL attributes", {
  node <- tags$div(id = "x", class = NULL)
  expect_equal(render(node), '<div id="x"></div>')
})

test_that("render collapses multi-value class attribute", {
  node <- tags$div(class = c("a", "b", "c"))
  expect_equal(render(node), '<div class="a b c"></div>')
})

# -- render.default --------------------------------------------------------

test_that("render.default escapes text", {
  expect_equal(render("a<b"), "a&lt;b")
})

test_that("render.default coerces to character", {
  expect_equal(render(42), "42")
})

test_that("render.default handles empty string", {
  expect_equal(render(""), "")
})

# -- render.list -----------------------------------------------------------

test_that("render.list concatenates rendered elements", {
  nodes <- list(tags$p("a"), tags$p("b"))
  expect_equal(render(nodes), "<p>a</p><p>b</p>")
})

test_that("render.list handles mixed tags and text", {
  nodes <- list("hello ", tags$strong("world"))
  expect_equal(render(nodes), "hello <strong>world</strong>")
})

test_that("render.list handles empty list", {
  expect_equal(render(list()), "")
})

test_that("render.list handles single element", {
  nodes <- list(tags$div("one"))
  expect_equal(render(nodes), "<div>one</div>")
})

# -- render with file (file output) ----------------------------------------

test_that("render writes tag to a file path and returns invisibly", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  node <- tags$div("hello")
  result <- withVisible(render(node, file = tmp))

  expect_false(result$visible)
  expect_equal(result$value, "<div>hello</div>")
  expect_equal(readLines(tmp, warn = FALSE), "<div>hello</div>")
})

test_that("render writes void element to a file path", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  node <- tags$br()
  render(node, file = tmp)

  expect_equal(readLines(tmp, warn = FALSE), "<br />")
})

test_that("render writes nested HTML to a file path", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  page <- tags$html(
    tags$head(tags$title("Test")),
    tags$body(tags$h1("Hello"))
  )
  render(page, file = tmp)

  expect_equal(
    readLines(tmp, warn = FALSE),
    "<html><head><title>Test</title></head><body><h1>Hello</h1></body></html>"
  )
})

test_that("render.default writes text to a file path", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  result <- withVisible(render("a<b", file = tmp))

  expect_false(result$visible)
  expect_equal(result$value, "a&lt;b")
  expect_equal(readLines(tmp, warn = FALSE), "a&lt;b")
})

test_that("render.list writes concatenated HTML to a file path", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  nodes <- list(tags$p("a"), tags$p("b"))
  result <- withVisible(render(nodes, file = tmp))

  expect_false(result$visible)
  expect_equal(result$value, "<p>a</p><p>b</p>")
  expect_equal(readLines(tmp, warn = FALSE), "<p>a</p><p>b</p>")
})

test_that("render writes to a connection object", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  con <- file(tmp, open = "w")
  on.exit(close(con), add = TRUE)

  node <- tags$div(class = "test", "content")
  render(node, file = con)

  # flush so readLines sees the output
  flush(con)
  expect_equal(readLines(tmp, warn = FALSE), '<div class="test">content</div>')
})

test_that("render with file = '' returns visibly (default behaviour)", {
  node <- tags$div("hello")
  result <- withVisible(render(node, file = ""))

  expect_true(result$visible)
  expect_equal(result$value, "<div>hello</div>")
})

test_that("render overwrites existing file content", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  render(tags$p("first"), file = tmp)
  render(tags$p("second"), file = tmp)

  expect_equal(readLines(tmp, warn = FALSE), "<p>second</p>")
})

# -- render with write_mode ------------------------------------------------

test_that("render overwrites file when write_mode = 'overwrite' explicitly", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  render(tags$p("first"), file = tmp)
  render(tags$p("second"), file = tmp, write_mode = "overwrite")

  expect_equal(readLines(tmp, warn = FALSE), "<p>second</p>")
})

test_that("render.hypertext.tag appends to file when write_mode = 'append'", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  render(tags$p("first"), file = tmp)
  render(tags$p("second"), file = tmp, write_mode = "append")

  expect_equal(
    readLines(tmp, warn = FALSE),
    "<p>first</p><p>second</p>"
  )
})

test_that("render.default appends to file when write_mode = 'append'", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  render("hello", file = tmp)
  render(" world", file = tmp, write_mode = "append")

  expect_equal(readLines(tmp, warn = FALSE), "hello world")
})

test_that("render.list appends to file when write_mode = 'append'", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  render(list(tags$p("a")), file = tmp)
  render(list(tags$p("b")), file = tmp, write_mode = "append")

  expect_equal(readLines(tmp, warn = FALSE), "<p>a</p><p>b</p>")
})

test_that("render appends across multiple calls with write_mode = 'append'", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  render(tags$h1("Title"), file = tmp)
  render(tags$p("para 1"), file = tmp, write_mode = "append")
  render(tags$p("para 2"), file = tmp, write_mode = "append")

  expect_equal(
    readLines(tmp, warn = FALSE),
    "<h1>Title</h1><p>para 1</p><p>para 2</p>"
  )
})

test_that("render errors on invalid write_mode value", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  expect_error(
    render(tags$p("x"), file = tmp, write_mode = "bogus"),
    "arg"
  )
})

test_that("render appends to connection object with write_mode = 'append'", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  con <- file(tmp, open = "w")
  on.exit(close(con), add = TRUE)

  render(tags$p("first"), file = con)
  render(tags$p("second"), file = con, write_mode = "append")

  flush(con)
  expect_equal(
    readLines(tmp, warn = FALSE),
    "<p>first</p><p>second</p>"
  )
})

test_that("write_mode is ignored when file = '' (returns visibly)", {
  node <- tags$div("hello")
  result <- withVisible(render(node, file = "", write_mode = "append"))

  expect_true(result$visible)
  expect_equal(result$value, "<div>hello</div>")
})

# -- tag() public API for custom elements ----------------------------------

test_that("tag() creates a hypertext.tag node", {
  node <- tag("my-element")
  expect_s3_class(node, "hypertext.tag")
  expect_equal(node$tag, "my-element")
})

test_that("tag() accepts named args as attributes", {
  node <- tag("calcite-action-bar", layout = "horizontal")
  expect_equal(node$attrs$layout, "horizontal")
})

test_that("tag() accepts unnamed args as children", {
  node <- tag("my-card", "hello", "world")
  expect_equal(node$children, list("hello", "world"))
})

test_that("tag() renders correctly with attributes and children", {
  html <- render(tag("calcite-action-bar", layout = "horizontal"))
  expect_equal(
    html,
    "<calcite-action-bar layout=\"horizontal\"></calcite-action-bar>"
  )
})

test_that("tag() supports nested custom tags", {
  node <- tag(
    "my-card",
    class = "shadow",
    tag("my-card-header", "Title"),
    tag("my-card-body", "Content")
  )
  html <- render(node)
  expect_equal(
    html,
    paste0(
      "<my-card class=\"shadow\">",
      "<my-card-header>Title</my-card-header>",
      "<my-card-body>Content</my-card-body>",
      "</my-card>"
    )
  )
})

test_that("tag() works with tag_type = 'void' (self-closing)", {
  node <- tag("my-icon", name = "home", tag_type = "void")
  expect_length(node$children, 0)
  html <- render(node)
  expect_equal(html, "<my-icon name=\"home\" />")
})

test_that("tag() tag_type = 'void' ignores children", {
  node <- tag("my-icon", "ignored child", name = "home", tag_type = "void")
  expect_length(node$children, 0)
})

test_that("tag() defaults to tag_type = 'normal'", {
  node <- tag("my-element")
  expect_equal(node$tag_type, "normal")
  html <- render(node)
  expect_equal(html, "<my-element></my-element>")
})

test_that("tag() rejects invalid tag_type", {
  expect_error(tag("my-el", tag_type = "bogus"), "arg")
})

test_that("tag() can mix with built-in tags", {
  html <- render(
    tags$div(
      class = "wrapper",
      tag("calcite-shell", tag("calcite-action-bar", layout = "horizontal"))
    )
  )
  expect_equal(
    html,
    paste0(
      "<div class=\"wrapper\">",
      "<calcite-shell>",
      "<calcite-action-bar layout=\"horizontal\">",
      "</calcite-action-bar>",
      "</calcite-shell>",
      "</div>"
    )
  )
})

test_that("tag() flattens list children like built-in tags", {
  items <- list("a", "b", "c")
  node <- tag("my-list", lapply(items, function(x) tag("my-item", x)))
  expect_length(node$children, 3)
  html <- render(node)
  expect_equal(
    html,
    paste0(
      "<my-list>",
      "<my-item>a</my-item>",
      "<my-item>b</my-item>",
      "<my-item>c</my-item>",
      "</my-list>"
    )
  )
})

test_that("tag() escapes attribute values", {
  html <- render(tag("my-el", title = "a \"quoted\" <value>"))
  expect_match(html, "&quot;")
  expect_match(html, "&lt;")
})

test_that("tag() supports boolean attributes", {
  html <- render(tag("my-el", disabled = TRUE))
  expect_equal(html, "<my-el disabled></my-el>")
})

test_that("tag() drops FALSE and NULL attributes", {
  html <- render(tag("my-el", hidden = FALSE, data = NULL))
  expect_equal(html, "<my-el></my-el>")
})

test_that("tag() drops NULL children", {
  node <- tag("p", "this", NULL)
  expect_equal(node$children, list("this"))
  expect_equal(render(node), "<p>this</p>")
})

test_that("tags$p drops NULL children", {
  html <- render(tags$p("this", NULL))
  expect_equal(html, "<p>this</p>")
})

test_that("tag() handles only NULL children", {
  node <- tag("div", NULL)
  expect_equal(node$children, list())
  expect_equal(render(node), "<div></div>")
})

# -- trailing commas in tags$*() --------------------------------------------

test_that("trailing comma after single child in tags$*()", {
  node <- tags$p(
    "hello",
  )
  expect_s3_class(node, "hypertext.tag")
  expect_equal(node$children, list("hello"))
  expect_equal(render(node), "<p>hello</p>")
})

test_that("trailing comma after multiple children in tags$*()", {
  node <- tags$p(
    tags$span("a"),
    tags$span("b"),
  )
  expect_length(node$children, 2L)
  expect_equal(
    render(node),
    "<p><span>a</span><span>b</span></p>"
  )
})

test_that("trailing comma after single attribute in tags$*()", {
  node <- tags$div(
    class = "container",
  )
  expect_equal(node$attrs$class, "container")
  expect_length(node$children, 0L)
  expect_equal(render(node), '<div class="container"></div>')
})

test_that("trailing comma after children when attributes present", {
  node <- tags$div(
    class = "wrapper",
    tags$p("content"),
  )
  expect_equal(node$attrs$class, "wrapper")
  expect_length(node$children, 1L)
  expect_equal(
    render(node),
    '<div class="wrapper"><p>content</p></div>'
  )
})

test_that("trailing comma in void tag", {
  node <- tags$input(
    type = "text",
    name = "field",
  )
  expect_equal(node$tag_type, "void")
  expect_equal(
    render(node),
    '<input type="text" name="field" />'
  )
})

# -- trailing commas in tag_list() ------------------------------------------

test_that("trailing comma in tag_list() with single element", {
  tl <- tag_list(
    tags$p("one"),
  )
  expect_s3_class(tl, "hypertext.tag.list")
  expect_length(tl, 1L)
  expect_equal(render(tl), "<p>one</p>")
})

test_that("trailing comma in tag_list() with multiple elements", {
  tl <- tag_list(
    tags$h1("title"),
    tags$p("body"),
  )
  expect_length(tl, 2L)
  expect_equal(render(tl), "<h1>title</h1><p>body</p>")
})

test_that("trailing comma in tag_list() with mixed tags and text", {
  tl <- tag_list(
    "hello ",
    tags$strong("world"),
  )
  expect_length(tl, 2L)
  expect_equal(render(tl), "hello <strong>world</strong>")
})

# -- nested tags with trailing commas ---------------------------------------

test_that("trailing comma in both outer and inner nested tags", {
  node <- tags$div(
    tags$p(
      tags$span("inner"),
    ),
  )
  expect_equal(
    render(node),
    "<div><p><span>inner</span></p></div>"
  )
})

test_that("trailing comma only in outer tag", {
  node <- tags$div(
    tags$p(tags$span("inner")),
  )
  expect_equal(
    render(node),
    "<div><p><span>inner</span></p></div>"
  )
})

test_that("trailing comma only in inner tag", {
  node <- tags$div(
    tags$p(
      tags$span("inner"),
    )
  )
  expect_equal(
    render(node),
    "<div><p><span>inner</span></p></div>"
  )
})

test_that("deeply nested tags with trailing commas at every level", {
  node <- tags$div(
    tags$section(
      tags$article(
        tags$p(
          tags$span("deep"),
        ),
      ),
    ),
  )
  expect_equal(
    render(node),
    paste0(
      "<div><section><article>",
      "<p><span>deep</span></p>",
      "</article></section></div>"
    )
  )
})

# -- trailing commas referencing outer scope variables ----------------------

test_that("trailing comma after variable indexing", {
  x <- c("alpha", "beta", "gamma")
  node <- tags$p(
    x[1],
  )
  expect_equal(node$children, list("alpha"))
  expect_equal(render(node), "<p>alpha</p>")
})

test_that("trailing comma after multiple variable references", {
  x <- c("hello", "world")
  node <- tag_list(
    tags$div(
      tags$p(
        x[1],
        " ",
      ),
      tags$p(
        x[2],
      ),
    ),
  )
  expect_equal(render(node), "<div><p>hello </p><p>world</p></div>")
})

test_that("trailing comma after lapply referencing outer variable", {
  items <- c("a", "b", "c")
  node <- tags$ul(
    lapply(
      X = items,
      FUN = function(x) {
        tags$li(
          paste("item:", x),
        )
      }
    ),
  )
  expect_length(node$children, 3L)
  expect_equal(
    render(node),
    "<ul><li>item: a</li><li>item: b</li><li>item: c</li></ul>"
  )
})

test_that("tag_list with lapply and trailing comma", {
  items <- c("a", "b", "c")
  tl <- tag_list(
    tags$h3("a page"),
    lapply(items, tags$p),
  )
  expect_length(tl, 2L)
  expect_equal(
    render(tl),
    "<h3>a page</h3><p>a</p><p>b</p><p>c</p>"
  )
})

test_that("nested tags referencing outer variable with trailing commas", {
  x <- c("a", "b", "c")
  node <- tag_list(
    tags$h3("heading"),
    tags$p(x[1], ),
    lapply(x, tags$span),
  )
  expect_length(node, 3L)
  expect_equal(
    render(node),
    paste0(
      "<h3>heading</h3>",
      "<p>a</p>",
      "<span>a</span><span>b</span><span>c</span>"
    )
  )
})

test_that("trailing comma with variable used as attribute value", {
  cls <- "highlight"
  node <- tags$div(
    class = cls,
    tags$p("text"),
  )
  expect_equal(node$attrs$class, "highlight")
  expect_equal(
    render(node),
    '<div class="highlight"><p>text</p></div>'
  )
})

test_that("trailing comma with computed expression as child", {
  vals <- 1:3
  node <- tags$p(
    paste(vals, collapse = ", "),
  )
  expect_equal(render(node), "<p>1, 2, 3</p>")
})

test_that("trailing comma after variable in nested tag inside tag_list", {
  title <- "My Page"
  items <- c("first", "second")
  tl <- tag_list(
    tags$h1(title, ),
    tags$ul(
      lapply(items, function(item) {
        tags$li(item, )
      }),
    ),
  )
  expect_equal(
    render(tl),
    paste0(
      "<h1>My Page</h1>",
      "<ul><li>first</li><li>second</li></ul>"
    )
  )
})

# -- trailing commas with NULL handling -------------------------------------

test_that("trailing comma with NULL child", {
  node <- tags$p(
    "text",
    NULL,
  )
  expect_equal(node$children, list("text"))
  expect_equal(render(node), "<p>text</p>")
})

test_that("trailing comma in tag_list with NULL element", {
  tl <- tag_list(
    tags$p("a"),
    NULL,
  )
  expect_length(tl, 1L)
  expect_equal(render(tl), "<p>a</p>")
})

test_that("trailing comma with only NULL in tags$*()", {
  node <- tags$div(
    NULL,
  )
  expect_equal(node$children, list())
  expect_equal(render(node), "<div></div>")
})

# -- trailing commas with raw_html -----------------------------------------

test_that("trailing comma after raw_html child", {
  node <- tags$div(
    raw_html("<em>raw</em>"),
  )
  expect_equal(render(node), "<div><em>raw</em></div>")
})

test_that("trailing comma in tag_list with raw_html", {
  tl <- tag_list(
    tags$p("text"),
    raw_html("<hr />"),
  )
  expect_equal(render(tl), "<p>text</p><hr />")
})

# -- trailing commas with boolean and special attributes --------------------

test_that("trailing comma after boolean attribute", {
  node <- tags$input(
    type = "checkbox",
    checked = TRUE,
  )
  html <- render(node)
  expect_match(html, "checked")
  expect_match(html, 'type="checkbox"')
})

test_that("trailing comma with FALSE and NULL attributes", {
  node <- tags$div(
    hidden = FALSE,
    data = NULL,
  )
  expect_equal(render(node), "<div></div>")
})

# -- trailing commas with numeric children ----------------------------------

test_that("trailing comma after numeric child", {
  node <- tags$span(
    42,
  )
  expect_equal(render(node), "<span>42</span>")
})

test_that("trailing comma with mixed numeric and text", {
  node <- tags$p(
    "Score: ",
    100,
  )
  expect_equal(render(node), "<p>Score: 100</p>")
})

# -- trailing commas in realistic page structure ----------------------------

test_that("realistic page with trailing commas throughout", {
  items <- c("Home", "About", "Contact")
  page <- tags$html(
    tags$head(
      tags$title("My Site"),
      tags$meta(charset = "utf-8", ),
    ),
    tags$body(
      tags$nav(
        class = "main-nav",
        tags$ul(
          lapply(items, function(item) {
            tags$li(
              tags$a(href = "#", item, ),
            )
          }),
        ),
      ),
      tags$main(
        tags$h1("Welcome", ),
        tags$p("Hello world", ),
      ),
    ),
  )
  html <- render(page)
  expect_match(html, "^<html>")
  expect_match(html, "</html>$")
  expect_match(html, "<title>My Site</title>")
  expect_match(html, '<meta charset="utf-8" />')
  expect_match(html, '<nav class="main-nav">')
  expect_match(html, "<li><a href=\"#\">Home</a></li>")
  expect_match(html, "<li><a href=\"#\">About</a></li>")
  expect_match(html, "<li><a href=\"#\">Contact</a></li>")
  expect_match(html, "<h1>Welcome</h1>")
  expect_match(html, "<p>Hello world</p>")
})

test_that("form with trailing commas", {
  form <- tags$form(
    action = "/submit",
    method = "post",
    tags$label(`for` = "email", "Email:", ),
    tags$input(type = "email", id = "email", name = "email", ),
    tags$button(type = "submit", "Send", ),
  )
  html <- render(form)
  expect_match(html, '<form action="/submit" method="post">')
  expect_match(html, '<label for="email">Email:</label>')
  expect_match(html, '<input type="email" id="email" name="email" />')
  expect_match(html, "<button type=\"submit\">Send</button>")
})

test_that("table built with trailing commas and lapply", {
  headers <- c("Name", "Age")
  rows <- list(
    c("Alice", "30"),
    c("Bob", "25")
  )
  tbl <- tags$table(
    class = "data",
    tags$thead(
      tags$tr(
        lapply(headers, tags$th),
      ),
    ),
    tags$tbody(
      lapply(rows, function(row) {
        tags$tr(
          lapply(row, tags$td),
        )
      }),
    ),
  )
  html <- render(tbl)
  expect_match(html, '<table class="data">')
  expect_match(html, "<th>Name</th><th>Age</th>")
  expect_match(html, "<td>Alice</td><td>30</td>")
  expect_match(html, "<td>Bob</td><td>25</td>")
})

# -- trailing commas with tag_list as child of tag --------------------------

test_that("trailing comma with tag_list as child of tag", {
  tl <- tag_list(
    tags$li("one"),
    tags$li("two"),
  )
  node <- tags$ul(
    tl,
  )
  expect_equal(render(node), "<ul><li>one</li><li>two</li></ul>")
})

test_that("trailing comma with tag_list mixed with other children", {
  tl <- tag_list(
    tags$li("b"),
    tags$li("c"),
  )
  node <- tags$ul(
    tags$li("a"),
    tl,
    tags$li("d"),
  )
  expect_equal(
    render(node),
    "<ul><li>a</li><li>b</li><li>c</li><li>d</li></ul>"
  )
})

# -- trailing commas with multiple class values ----------------------------

test_that("trailing comma with collapsed class vector", {
  node <- tags$div(
    class = c("container", "mx-auto", "p-4"),
    "content",
  )
  expect_equal(
    render(node),
    '<div class="container mx-auto p-4">content</div>'
  )
})

# -- trailing commas with conditional (NULL) children ----------------------

test_that("trailing comma with conditional NULL children from expressions", {
  show_extra <- FALSE
  node <- tags$div(
    tags$p("always"),
    if (show_extra) tags$p("extra"),
  )
  expect_length(node$children, 1L)
  expect_equal(render(node), "<div><p>always</p></div>")
})

test_that("trailing comma with conditional TRUE children from expressions", {
  show_extra <- TRUE
  node <- tags$div(
    tags$p("always"),
    if (show_extra) tags$p("extra"),
  )
  expect_length(node$children, 2L)
  expect_equal(
    render(node),
    "<div><p>always</p><p>extra</p></div>"
  )
})

test_that("trailing comma in tag_list with conditional children", {
  show_nav <- FALSE
  tl <- tag_list(
    tags$h1("Title"),
    if (show_nav) tags$nav("nav content"),
    tags$main("main content"),
  )
  # NULL from if() is filtered out by tag_list
  expect_length(tl, 2L)
  expect_equal(
    render(tl),
    "<h1>Title</h1><main>main content</main>"
  )
})

test_that("trailing comma works with forwarded ellipsis in functions", {
  page <- function(...) {
    content <- tag_list(
      doctype(),
      tags$html(),
      tags$head(
        tags$meta(
          charset = "utf-8"
        )
      ),
      tags$body(
        tags$main(
          class = "main",
          ...,
        ),
      ),
    )

    render(content)
  }

  home_ui <- function() {
    content <- tags$div(
      tags$h3("hello, world"),
      tags$p("this is a test"),
    )

    page(content)
  }

  expect_equal(
    home_ui(),
    "<!DOCTYPE html><html></html><head><meta charset=\"utf-8\" /></head><body><main class=\"main\"><div><h3>hello, world</h3><p>this is a test</p></div></main></body>"
  )
})

test_that(".flatten_children returns empty list for empty input", {
  expect_equal(hypertext:::.flatten_children(list()), list())
})

test_that(".flatten_children keeps non-list elements as-is", {
  result <- hypertext:::.flatten_children(list("a", "b"))
  expect_equal(result, list("a", "b"))
})

test_that(".flatten_children unpacks plain lists", {
  result <- hypertext:::.flatten_children(list(list("a", "b"), "c"))
  expect_equal(result, list("a", "b", "c"))
})

test_that(".flatten_children recursively unpacks nested lists", {
  result <- hypertext:::.flatten_children(list(list(list("deep"))))
  expect_equal(result, list("deep"))
})

test_that(".flatten_children preserves hypertext.tag objects", {
  tag_node <- hypertext::tag("span", "text")
  result <- hypertext:::.flatten_children(list(tag_node))
  expect_length(result, 1)
  expect_s3_class(result[[1]], "hypertext.tag")
})

test_that(".flatten_children unpacks lists but preserves tags inside", {
  tag_node <- hypertext::tag("li", "item")
  result <- hypertext:::.flatten_children(list(list(tag_node, "text")))
  expect_length(result, 2)
  expect_s3_class(result[[1]], "hypertext.tag")
  expect_equal(result[[2]], "text")
})

test_that(".flatten_children handles lapply-style output", {
  items <- c("a", "b", "c")
  tag_list <- lapply(items, function(x) hypertext::tag("li", x))
  result <- hypertext:::.flatten_children(list(tag_list))
  expect_length(result, 3)
  for (i in seq_along(result)) {
    expect_s3_class(result[[i]], "hypertext.tag")
  }
})

test_that(".flatten_children preserves numeric elements", {
  result <- hypertext:::.flatten_children(list(1, 2, 3))
  expect_equal(result, list(1, 2, 3))
})

test_that(".flatten_children drops NULL elements", {
  result <- hypertext:::.flatten_children(list("a", NULL, "b"))
  expect_equal(result, list("a", "b"))
})

test_that(".flatten_children drops NULL inside nested lists", {
  result <- hypertext:::.flatten_children(list(list("a", NULL, "b")))
  expect_equal(result, list("a", "b"))
})

test_that(".flatten_children handles all-NULL input", {
  result <- hypertext:::.flatten_children(list(NULL, NULL))
  expect_equal(result, list())
})

# -- tag_list constructor ---------------------------------------------------

test_that("tag_list returns object with correct classes", {
  tl <- tag_list(tags$p("a"))
  expect_s3_class(tl, "hypertext.tag.list")
  expect_s3_class(tl, "list")
  expect_equal(class(tl), c("hypertext.tag.list", "list"))
})

test_that("tag_list stores children as list elements", {
  tl <- tag_list(tags$p("a"), tags$p("b"))
  expect_length(tl, 2L)
  expect_s3_class(tl[[1]], "hypertext.tag")
  expect_s3_class(tl[[2]], "hypertext.tag")
})

test_that("tag_list with no arguments produces empty tag list", {
  tl <- tag_list()
  expect_s3_class(tl, "hypertext.tag.list")
  expect_length(tl, 0L)
})

test_that("tag_list drops NULL elements", {
  tl <- tag_list(tags$p("a"), NULL, tags$p("b"))
  expect_s3_class(tl, "hypertext.tag.list")
  expect_length(tl, 2L)
  expect_equal(render(tl), "<p>a</p><p>b</p>")
})

test_that("tag_list with only NULL produces empty tag list", {
  tl <- tag_list(NULL)
  expect_s3_class(tl, "hypertext.tag.list")
  expect_length(tl, 0L)
  expect_equal(render(tl), "")
})

test_that("tag_list accepts mixed tags and text", {
  tl <- tag_list("hello ", tags$strong("world"))
  expect_length(tl, 2L)
  expect_equal(tl[[1]], "hello ")
  expect_s3_class(tl[[2]], "hypertext.tag")
})

test_that("tag_list accepts a single element", {
  tl <- tag_list(tags$div("only"))
  expect_length(tl, 1L)
})

# -- rendering tag_list ----------------------------------------------------

test_that("render concatenates tag_list children", {
  tl <- tag_list(tags$p("a"), tags$p("b"))
  expect_equal(render(tl), "<p>a</p><p>b</p>")
})

test_that("render handles empty tag_list", {
  tl <- tag_list()
  expect_equal(render(tl), "")
})

test_that("render handles tag_list with mixed tags and text", {
  tl <- tag_list("hello ", tags$strong("world"))
  expect_equal(render(tl), "hello <strong>world</strong>")
})

test_that("render handles tag_list with single element", {
  tl <- tag_list(tags$div("one"))
  expect_equal(render(tl), "<div>one</div>")
})

test_that("render tag_list writes to file and returns invisibly", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  tl <- tag_list(tags$p("a"), tags$p("b"))
  result <- withVisible(render(tl, file = tmp))

  expect_false(result$visible)
  expect_equal(result$value, "<p>a</p><p>b</p>")
  expect_equal(readLines(tmp, warn = FALSE), "<p>a</p><p>b</p>")
})

test_that("render tag_list supports write_mode = 'append'", {
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp), add = TRUE)

  render(tags$h1("Title"), file = tmp)
  render(tag_list(tags$p("a"), tags$p("b")), file = tmp, write_mode = "append")

  expect_equal(
    readLines(tmp, warn = FALSE),
    "<h1>Title</h1><p>a</p><p>b</p>"
  )
})

test_that("render handles nested tag_list", {
  inner <- tag_list(tags$li("a"), tags$li("b"))
  outer <- tags$ul(inner)
  expect_equal(render(outer), "<ul><li>a</li><li>b</li></ul>")
})

# -- print.hypertext.tag.list ----------------------------------------------

test_that("print.hypertext.tag.list outputs rendered HTML with newline", {
  tl <- tag_list(tags$p("a"), tags$p("b"))
  expect_output(print(tl), "<p>a</p><p>b</p>")
})

test_that("print.hypertext.tag.list returns invisibly", {
  tl <- tag_list(tags$p("a"))
  result <- withVisible(print(tl))
  expect_false(result$visible)
})

# -- .flatten_children preserves tag_list ----------------------------------

test_that(".flatten_children does not unwrap tag_list", {
  tl <- tag_list(tags$li("a"), tags$li("b"))
  result <- hypertext:::.flatten_children(list(tl))
  expect_length(result, 1L)
  expect_s3_class(result[[1]], "hypertext.tag.list")
})

test_that(".flatten_children still flattens plain lists", {
  plain <- list(tags$li("a"), tags$li("b"))
  result <- hypertext:::.flatten_children(list(plain))
  expect_length(result, 2L)
  expect_s3_class(result[[1]], "hypertext.tag")
  expect_s3_class(result[[2]], "hypertext.tag")
})

# -- tag_list as child of a tag --------------------------------------------

test_that("tag_list as child of tag renders its children inline", {
  tl <- tag_list(tags$li("one"), tags$li("two"))
  node <- tags$ul(tl)
  expect_equal(render(node), "<ul><li>one</li><li>two</li></ul>")
})

test_that("tag_list mixed with other children in a tag", {
  tl <- tag_list(tags$li("b"), tags$li("c"))
  node <- tags$ul(tags$li("a"), tl, tags$li("d"))
  expect_equal(
    render(node),
    "<ul><li>a</li><li>b</li><li>c</li><li>d</li></ul>"
  )
})

# -- .make_tag --------------------------------

test_that(".make_tag returns a function", {
  fn <- hypertext:::.make_tag("div")
  expect_true(is.function(fn))
})

test_that(".make_tag creates correct tag nodes", {
  fn <- hypertext:::.make_tag("section")
  node <- fn(class = "main", "content")
  expect_s3_class(node, "hypertext.tag")
  expect_equal(node$tag, "section")
  expect_equal(node$attrs$class, "main")
  expect_equal(node$children[[1]], "content")
})

test_that(".make_tag captures tag_name correctly via force()", {
  # Create functions in a loop to test closure capture
  fns <- list()
  for (name in c("div", "span", "p")) {
    fns[[name]] <- hypertext:::.make_tag(name)
  }
  expect_equal(fns$div()$tag, "div")
  expect_equal(fns$span()$tag, "span")
  expect_equal(fns$p()$tag, "p")
})

# -- the `tags` list -------------------------------------------------------

test_that("tags is a named list", {
  expect_true(is.list(tags))
  expect_true(!is.null(names(tags)))
})

test_that("all entries in tags are functions", {
  for (nm in names(tags)) {
    expect_true(is.function(tags[[nm]]), info = paste("tags$", nm))
  }
})

test_that("tags contains common HTML elements", {
  common <- c(
    "div",
    "span",
    "p",
    "a",
    "h1",
    "h2",
    "h3",
    "ul",
    "ol",
    "li",
    "table",
    "tr",
    "td",
    "th",
    "form",
    "input",
    "button",
    "select",
    "img",
    "br",
    "hr",
    "head",
    "body",
    "html",
    "script",
    "style"
  )
  for (el in common) {
    expect_true(el %in% names(tags), info = paste("missing:", el))
  }
})

test_that("tags has 173 elements", {
  expect_equal(length(tags), 173)
})

test_that("void tags produce self-closing HTML", {
  void_names <- c(
    "area",
    "base",
    "br",
    "col",
    "embed",
    "hr",
    "img",
    "input",
    "link",
    "meta",
    "param",
    "source",
    "track",
    "wbr"
  )
  for (nm in void_names) {
    html <- render(tags[[nm]]())
    expect_match(html, "/>$", info = paste(nm, "should be self-closing"))
  }
})

test_that("non-void tags produce opening and closing tags", {
  non_void <- c("div", "span", "p", "a", "ul", "li", "table")
  for (nm in non_void) {
    html <- render(tags[[nm]]())
    expect_match(
      html,
      paste0("^<", nm, "><\\/", nm, ">$"),
      info = paste(nm, "should have closing tag")
    )
  }
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tags.R
\name{tag_list}
\alias{tag_list}
\title{Create a tag list (sibling container)}
\usage{
tag_list(...)
}
\arguments{
\item{...}{Tags, text, or other renderable objects /// Optional.}
}
\value{
A list of class \code{"hypertext.tag.list"}.
}
\description{
Useful when you want to collect sibling nodes into a single
object without wrapping them in a parent element.
}
\examples{
tl <- tag_list(tags$p("one"), tags$p("two"))
render(tl)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tags.R
\docType{data}
\name{tags}
\alias{tags}
\title{HTML tag functions}
\format{
An object of class \code{list} of length 115.
}
\usage{
tags
}
\description{
A named list of functions, one per HTML5 element. Access individual
tags with \code{tags$div(...)}, \code{tags$p(...)}, etc.
}
\details{
Named arguments become HTML attributes; unnamed arguments become
child nodes or text content.
}
\section{Boolean / valueless attributes}{

Pass \code{NA} or \code{TRUE} as the attribute value to produce a valueless
attribute (e.g. \code{disabled}). \code{FALSE} or \code{NULL} suppresses the
attribute entirely.
}

\section{Multi-value attributes}{

Character vectors of length > 1 are collapsed with a space, so
\code{class = c("a", "b")} renders as \code{class="a b"}.
}

\examples{
tags$p(class = "lead", "Hello, world!")
render(tags$div(id = "app", tags$h1("Title")))
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tags.R
\name{doctype}
\alias{doctype}
\title{Render the \verb{<!DOCTYPE html>} declaration}
\usage{
doctype()
}
\value{
A \code{"hypertext.raw"} object containing \verb{<!DOCTYPE html>}.
}
\description{
Convenience wrapper around \code{\link[=raw_html]{raw_html()}} that returns the
HTML5 document-type declaration. Useful when building a full page.
}
\examples{
page <- tag_list(
  doctype(),
  tags$html(
    tags$head(tags$title("Home")),
    tags$body(tags$h1("Welcome"))
  )
)
render(page)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tags.R
\name{tag}
\alias{tag}
\title{Create an HTML element node}
\usage{
tag(tag_name, ..., tag_type = c("normal", "void"))
}
\arguments{
\item{tag_name}{String /// Required.
The HTML element name.}

\item{...}{Attributes (named) and children (unnamed) /// Optional.}

\item{tag_type}{String /// Optional.
Either \code{"normal"} (default) for a standard
element with children, or \code{"void"} for a self-closing element
whose children are ignored.}
}
\value{
List of class \code{"hypertext.tag"} with components \code{tag}, \code{attrs},
\code{children}, and \code{tag_type}.
}
\description{
Low-level constructor. Named arguments become attributes; unnamed
arguments become children (text or nested nodes).
}
\examples{
# web component
tag(tag_name = "calcite-action-bar", layout = "horizontal")

# custom element with children
tag(
  tag_name = "my-card",
  class = "shadow",
  tag(tag_name = "my-card-header", "Title"),
  tag(tag_name = "my-card-body", "Content")
)

# custom void element
tag(tag_name = "my-icon", name = "home", tag_type = "void")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tags.R
\name{render}
\alias{render}
\alias{render.hypertext.tag}
\alias{render.default}
\alias{render.hypertext.raw}
\alias{render.list}
\title{Render an HTML node tree to a character string}
\usage{
render(x, ...)

\method{render}{hypertext.tag}(x, file = "", write_mode = c("overwrite", "append"), ...)

\method{render}{default}(x, file = "", write_mode = c("overwrite", "append"), ...)

\method{render}{hypertext.raw}(x, file = "", write_mode = c("overwrite", "append"), ...)

\method{render}{list}(x, file = "", write_mode = c("overwrite", "append"), ...)
}
\arguments{
\item{x}{\code{hypertext.tag} object, a string, or a list of these /// Required.}

\item{...}{Further arguments passed from or to other methods.}

\item{file}{String /// Optional.
Path to file to print to.}

\item{write_mode}{String /// Optional.
Either "overwrite" (default) which overwrites the contents
of \code{file}, or "append" which appends the HTML string to
\code{file}.}
}
\value{
A single character string of HTML.
}
\description{
Converts a \code{hypertext.tag} object (and all its descendants) into an HTML
string. Text children are escaped; nested \code{hypertext.tag} children are
rendered recursively.
}
\details{
When \code{file} is provided, the rendered HTML is written to the specified
file via \code{\link[=cat]{cat()}} and the HTML string is returned
invisibly.
}
\examples{
page <- tags$html(
  tags$head(
    tags$title("Home")
  ),
  tags$body(
    tags$h1("Welcome")
  )
)

# return HTML as a string:
render(page)

\dontrun{
  # write to a file:
  render(page, file = "index.html")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tags.R
\name{raw_html}
\alias{raw_html}
\title{Mark a string as raw HTML}
\usage{
raw_html(...)
}
\arguments{
\item{...}{Character strings of raw HTML /// Optional.}
}
\value{
A character vector of class \code{"hypertext.raw"}.
}
\description{
Wraps one or more character strings so that \code{\link[=render]{render()}} outputs them
verbatim, \strong{without} escaping HTML special characters.
}
\details{
This is useful for injecting pre-rendered HTML, inline
\verb{<script>}/\verb{<style>} blocks, SVG markup, or any content that should
not be entity-escaped.
}
\examples{
raw_html("<script>alert('hi')</script>")
render(tags$div(raw_html("<em>already formatted</em>")))
}
